#!/usr/bin/env python3
# coding=utf8

"""
Botan CI check installation script
This script is used to validate the results of `make install`

(C) 2020 Jack Lloyd, René Meusel, Hannes Rantzsch
(C) 2023 René Meusel

Botan is released under the Simplified BSD License (see license.txt)
"""

import os
import sys
import json
import re
import subprocess

def verify_library(build_config):
    lib_dir = build_config['libdir']
    if not os.path.isdir(lib_dir):
        print('Error: libdir "%s" is not a directory' % lib_dir)
        return False

    found_libs = set([])

    major_version = int(build_config["version_major"])

    if build_config['compiler'] == 'msvc':
        expected_lib_format = r'^botan-%d\.(dll|lib)$' % (major_version)
    elif build_config['os'] == 'macos':
        expected_lib_format = r'^libbotan-%d\.(a|dylib)$' % (major_version)
    else:
        expected_lib_format = r'^libbotan-%d\.(a|so)$' % (major_version)

    lib_re = re.compile(expected_lib_format)

    # Unlike the include dir this may have other random libs in it
    for (_, _, filenames) in os.walk(lib_dir):
        for filename in filenames:
            if lib_re.match(filename) is not None:
                found_libs.add(filename)

    if len(found_libs) == 0:
        print("Could not find any libraries from us")
        return False

    # This should match up the count and names of libraries installed
    # vs the build configuration (eg static lib installed or not)

    return True

def verify_includes(build_config):
    include_dir = build_config['installed_include_dir']
    if not os.path.isdir(include_dir):
        print('Error: installed_include_dir "%s" is not a directory' % include_dir)
        return False

    expected_headers = set(build_config['public_headers'] + build_config['external_headers'])
    found_headers = set([])

    for (_, _, filenames) in os.walk(include_dir):
        for filename in filenames:
            found_headers.add(filename)

    if found_headers != expected_headers:
        missing = expected_headers - found_headers
        extra = found_headers - expected_headers

        if len(missing) > 0:
            print("Missing expected headers: %s" % (" ".join(sorted(missing))))

        if len(extra) > 0:
            print("Have unexpected headers: %s" % (" ".join(sorted(extra))))
        return False

    return True

def verify_cmake_package(build_config):
    if build_config['os'] not in ['windows', 'linux', 'macos']:
        return True # skip (e.g. for mingw)

    cmake_build_dir = os.path.join(build_config['abs_root_dir'], build_config['build_dir'], 'cmake_test')
    cmake_test_dir = os.path.join(build_config['abs_root_dir'], "src", "scripts", "ci", "cmake_tests")

    def cmake_preset():
        if build_config['os'] == 'windows':
            return 'windows_x86_64' if build_config['arch'] == 'x86_64' else 'windows_x86'
        return 'unix'

    def test_target():
        return 'test' if build_config['os'] != 'windows' else 'RUN_TESTS'

    disabled_module = build_config['disabled_mod_list'][0] if build_config['disabled_mod_list'] else None
    if not disabled_module:
        print("Not a single disabled module in this build to use for testing.") # just for good measure
        return False

    try:
        subprocess.run(["cmake", "--preset", cmake_preset(),
                                 "-B", cmake_build_dir,
                                 "-S", cmake_test_dir,
                                 "-DCMAKE_PREFIX_PATH=%s" % build_config['prefix'],
                                 "-DBOTAN_DISABLED_MODULE=%s" % disabled_module], check=True)
        subprocess.run(["cmake", "--build", cmake_build_dir, "--config", "Release"], check=True)
        subprocess.run(["cmake", "--build", cmake_build_dir, "--config", "Release", "--target", test_target()], check=True)
    except RuntimeError as e:
        print("Using the CMake package failed: %s" % str(e))
        return False

    return True

def main(args=None):
    if args is None:
        args = sys.argv

    if len(args) < 2:
        print("Usage: %s <build_config.json>" % args[0])
        return 1

    with open(os.path.join(args[1]), encoding='utf8') as f:
        build_config = json.load(f)

    install_prefix = build_config['prefix']

    if not os.path.isdir(install_prefix):
        print('Error: install_prefix "%s" is not a directory' % install_prefix)
        return 1

    if not verify_includes(build_config):
        return 1

    if not verify_library(build_config):
        return 1

    has_cmake = 'botan_cmake_config' in build_config and 'botan_cmake_version_config' in build_config
    if has_cmake and not verify_cmake_package(build_config):
        return 1

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2022,2023 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import subprocess
import sys
import json
import optparse # pylint: disable=deprecated-module
import os
import re
import multiprocessing
import time
import uuid
from multiprocessing.pool import ThreadPool

quick_checks = [
    '-clang-analyzer*', # has to be explicitly disabled
    'modernize-use-nullptr',
    'readability-braces-around-statements',
    'performance-unnecessary-value-param',
    '*-non-private-member-variables-in-classes',
]

enabled_checks = [
    'bugprone-*',
    'cert-*',
    'clang-analyzer-*',
    'cppcoreguidelines-*',
    'hicpp-*',
    'misc-*',
    'modernize-*',
    'performance-*',
    'portability-*',
    'readability-*',
]

# these are ones that we might want to be clean for in the future,
# but currently are not
disabled_needs_work = [
    '*-named-parameter',
    '*-member-init', # should definitely fix this one
    'bugprone-lambda-function-name', # should be an easy fix
    'bugprone-unchecked-optional-access', # clang-tidy seems buggy (many false positives)
    'bugprone-empty-catch',
    'cert-err58-cpp', # many false positives eg __m128i
    'cppcoreguidelines-avoid-const-or-ref-data-members',
    'cppcoreguidelines-init-variables',
    'cppcoreguidelines-owning-memory',
    'cppcoreguidelines-prefer-member-initializer',
    'cppcoreguidelines-slicing', # private->public key slicing
    'hicpp-explicit-conversions',
    'misc-const-correctness', # pretty noisy
    'misc-include-cleaner',
    'misc-redundant-expression', # BigInt seems to confuse clang-tidy
    'misc-misplaced-const',
    'misc-confusable-identifiers',
    'modernize-avoid-bind',
    'modernize-pass-by-value',
    'modernize-use-ranges', # limited by compiler support currently
    'performance-avoid-endl',
    'readability-convert-member-functions-to-static',
    'readability-implicit-bool-conversion',
    'readability-inconsistent-declaration-parameter-name', # should fix this, blocked by https://github.com/llvm/llvm-project/issues/60845
    'readability-qualified-auto',
    'readability-simplify-boolean-expr', # sometimes ok
    'readability-static-accessed-through-instance',
]

# these we are probably not interested in ever being clang-tidy clean for
disabled_not_interested = [
    '*-array-to-pointer-decay',
    '*-avoid-c-arrays',
    '*-else-after-return',
    '*-function-size',
    '*-magic-numbers', # can't stop the magic
    '*-narrowing-conversions',
    '*-no-array-decay',
    '*-use-auto', # not universally a good idea
    '*-use-emplace', # often less clear
    '*-deprecated-headers', # wrong for system headers like stdlib.h
    'bugprone-argument-comment',
    'bugprone-branch-clone', # doesn't interact well with feature macros
    'bugprone-easily-swappable-parameters',
    'bugprone-implicit-widening-of-multiplication-result',
    'bugprone-suspicious-stringview-data-usage', # triggers on every use of string_view::data ??
    'cppcoreguidelines-avoid-do-while',
    'cppcoreguidelines-non-private-member-variables-in-classes', # pk split keys
    'cppcoreguidelines-pro-bounds-pointer-arithmetic',
    'cppcoreguidelines-pro-bounds-constant-array-index',
    'cppcoreguidelines-pro-type-const-cast', # see above
    'cppcoreguidelines-pro-type-reinterpret-cast', # not possible thanks though
    'cppcoreguidelines-pro-type-vararg', # idiocy
    'hicpp-no-assembler',
    'hicpp-vararg', # idiocy
    'hicpp-signed-bitwise', # impossible to avoid in C/C++, int promotion rules :/
    'misc-no-recursion',
    'modernize-loop-convert', # sometimes very ugly
    'modernize-raw-string-literal', # usually less readable
    'modernize-use-trailing-return-type', # fine, but we're not using it everywhere
    'modernize-return-braced-init-list', # thanks I hate it
    'modernize-use-default-member-init',
    'modernize-use-designated-initializers',
    'modernize-use-nodiscard',
    'modernize-use-using', # fine not great
    'portability-simd-intrinsics',
    'readability-avoid-return-with-void-value',
    'readability-container-data-pointer',
    'readability-function-cognitive-complexity',
    'readability-identifier-length', # lol, lmao
    'readability-isolate-declaration',
    'readability-math-missing-parentheses',
    'readability-non-const-parameter',
    'readability-redundant-access-specifiers', # reneme likes doing this
    'readability-suspicious-call-argument',
    'readability-use-std-min-max',
    'readability-use-anyofallof', # not more readable
]

disabled_checks = disabled_needs_work + disabled_not_interested

def create_check_option(enabled, disabled):
    return ','.join(enabled) + ',' + ','.join(['-' + d for d in disabled])

def render_clang_tidy_file(target_dir, enabled, disabled):
    filepath = os.path.join(target_dir, '.clang-tidy')
    print(f'regenerating {filepath}')
    with open(filepath, "w", encoding="utf-8") as clang_tidy_file:
        clang_tidy_file.writelines([
            '---\n',
            f'# This file was automatically generated by {sys.argv[0]} --regenerate-inline-config-file\n',
            '#\n',
            '# All manual edits to this file will be lost. Edit the script\n',
            '# then regenerate this configuration file.\n',
            '\n',
            'Checks: >\n'] +
            [ f'    {check},\n' for check in enabled ] +
            [ f'    -{check},\n' for check in disabled] +
            ['---\n'])

def load_compile_commands(build_dir):
    compile_commands_file = os.path.join(build_dir, 'compile_commands.json')
    compile_commands = open(compile_commands_file, encoding='utf8').read()
    return (compile_commands_file, json.loads(compile_commands))

def run_command(cmdline):
    proc = subprocess.Popen(cmdline,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    (stdout, _stderr) = proc.communicate()

    stdout = stdout.decode('utf8')
    # stderr is discarded

    return stdout

def run_clang_tidy(compile_commands_file,
                   check_config,
                   source_file,
                   options):

    cmdline = ['clang-tidy',
               '--quiet',
               '-checks=%s' % (check_config),
               '-p', compile_commands_file]

    if options.fixit:
        cmdline.append('-fix')

    if options.export_fixes_dir:
        os.makedirs(options.export_fixes_dir, exist_ok=True)
        yml_file = os.path.join(options.export_fixes_dir, "%s.yml" % uuid.uuid4())
        cmdline.append("--export-fixes=%s" % yml_file)

    cmdline.append(source_file)

    stdout = run_command(cmdline)

    if options.verbose:
        print("Checked", source_file)
        sys.stdout.flush()
    if stdout != "":
        print(stdout)
        sys.stdout.flush()
        return False

    return True

def file_matches(file, args):
    if args is None or len(args) == 0:
        return True

    for arg in args:
        if file.find(arg) > 0:
            return True
    return False

def main(args = None): # pylint: disable=too-many-return-statements
    if args is None:
        args = sys.argv

    parser = optparse.OptionParser()

    parser.add_option('-j', '--jobs', action='store', type='int', default=0)
    parser.add_option('--verbose', action='store_true', default=False)
    parser.add_option('--fixit', action='store_true', default=False)
    parser.add_option('--build-dir', default='build')
    parser.add_option('--list-checks', action='store_true', default=False)
    parser.add_option('--regenerate-inline-config-file', action='store_true', default=False)
    parser.add_option('--fast-checks-only', action='store_true', default=False)
    parser.add_option('--only-changed-files', action='store_true', default=False)
    parser.add_option('--only-matching', metavar='REGEX', default='.*')
    parser.add_option('--take-file-list-from-stdin', action='store_true', default=False)
    parser.add_option('--export-fixes-dir', default=None)

    (options, args) = parser.parse_args(args)

    if len(args) > 1:
        print("ERROR: Unknown extra arguments to run_clang_tidy.py")
        return 1

    if options.only_changed_files and options.take_file_list_from_stdin:
        print("Cannot use both --only-changed-files and --take-file-list-from-stdin", file=sys.stderr)
        return 1

    files_to_check = []
    if options.only_changed_files:
        changes = run_command(['git', 'diff', '--name-only', '-r', 'master'])

        files_to_check = []
        for file in changes.split():
            if file.endswith('.cpp') or file.endswith('.h'):
                files_to_check.append(os.path.basename(file))

        if len(files_to_check) == 0:
            print("No C++ files were modified vs master, skipping clang-tidy checks")
            return 0
    elif options.take_file_list_from_stdin:
        for line in sys.stdin:
            file = os.path.basename(line.strip())
            if file.endswith('.cpp') or file.endswith('.h'):
                files_to_check.append(file)

        if len(files_to_check) == 0:
            print("No C++ files were provided on stdin, skipping clang-tidy checks")
            return 0

    jobs = options.jobs
    if jobs == 0:
        jobs = multiprocessing.cpu_count() + 1

    (compile_commands_file, compile_commands) = load_compile_commands(options.build_dir)

    if options.fast_checks_only:
        check_config = ','.join(quick_checks)
    else:
        check_config = create_check_option(enabled_checks, disabled_checks)

    if options.list_checks:
        print(run_command(['clang-tidy', '-list-checks', '-checks', check_config]), end='')
        return 0

    if options.regenerate_inline_config_file:
        render_clang_tidy_file('src', enabled_checks, disabled_checks)
        return 0

    pool = ThreadPool(jobs)

    start_time = time.time()
    files_checked = 0

    file_matcher = re.compile(options.only_matching)

    results = []
    for info in compile_commands:
        file = info['file']

        if len(files_to_check) > 0 and os.path.basename(file) not in files_to_check:
            continue

        if file_matcher.search(file) is None:
            continue

        files_checked += 1
        results.append(pool.apply_async(
            run_clang_tidy,
            (compile_commands_file,
             check_config,
             file,
             options)))

    fail_cnt = 0
    for result in results:
        if not result.get():
            fail_cnt += 1

    time_consumed = time.time() - start_time

    print("Checked %d files in %d seconds" % (files_checked, time_consumed))

    if fail_cnt == 0:
        return 0
    else:
        print("Found clang-tidy errors in %d files" % (fail_cnt))
        return 1

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

#
# Strips the KAT harness in the Kyber reference implementation down to a less
# space consuming version. This script was used to generate:
#   * `src/tests/data/pubkey/kyber_kat.vec` and
#   * `src/tests/data/pubkey/ml_kem_ipd.vec`
# test data from the *.rsp files in the reference implementation repository.
#
# See here: https://github.com/pq-crystals/kyber
#
# NOTE! The reference implementation does not distinguish between vanilla
#       Kyber and the 90s mode when generating the *.rsp files. Run this
#       script to retrofit the distinction:
#
#       rm -f *.rsp
#       mkdir KATs
#       for gen in PQCgenKAT_kem1024 PQCgenKAT_kem1024-90s PQCgenKAT_kem512 PQCgenKAT_kem512-90s PQCgenKAT_kem768 PQCgenKAT_kem768-90s; do
#         ./$gen
#         mv *.rsp KATs/${gen}.rsp
#       done
#
# (C) 2023 Jack Lloyd
# (C) 2023 René Meusel - Rohde & Schwarz Cybersecurity
# (C) 2024 Amos Treiber, René Meusel - Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

import sys
import hashlib
import binascii
import os
import argparse

class KatReader:
    def __init__(self, file):
        self.file = file

    def next_value(self):
        while True:
            line = self.file.readline()

            if line == "":
                return (None, None)

            if line.startswith('#') or line == "\n":
                continue

            key, val = line.strip().split(' = ')

            return (key, val)

    def read_kats(self):
        kat = {}

        while True:
            key, val = self.next_value()

            if key == None:
                return # eof

            if key in ['msg']:
                continue

            if key not in ['count', 'z', 'd', 'seed', 'pk', 'sk', 'ct', 'ss', 'ct_n', 'ss_n']:
                raise Exception("Unknown key %s" % (key))

            if key == 'count':
                kat[key] = int(val)
            else:
                kat[key] = val

            if key == 'ss':
                yield kat
                kat = {}

def shake_256_16(v):
    # v is assumed to be hex
    h = hashlib.shake_256()
    h.update(binascii.unhexlify(v))
    return h.hexdigest(16)

def sha256_16(v):
    # v is assumed to be hex
    h = hashlib.sha256()
    h.update(binascii.unhexlify(v))
    return h.hexdigest()[:32]

def compress_kat(kat, mode):
    first = kat['count'] == 0
    del kat['count']

    hash_fn = sha256_16 if '90s' in mode else shake_256_16

    # For ML-KEM we use the private seed as the private key's storage format
    if mode == "ML-KEM":
        kat['sk'] = kat['d'] + kat['z']
        del kat['d']
        del kat['z']

    # rename keys and hash large values to reduce size of KAT vectors
    kat['Seed'] = kat.pop('seed')
    kat['SS']   = kat.pop('ss')
    kat['PK']   = hash_fn(kat.pop('pk'))
    kat['SK']   = hash_fn(kat.pop('sk'))
    kat['CT']   = hash_fn(kat.pop('ct'))

    if mode == "ML-KEM":
        kat['CT_N'] = kat.pop('ct_n')
        kat['SS_N'] = kat.pop('ss_n')

    return kat


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+", help="Input files")
    parser.add_argument("--kyber-r3", action="store_true", help="Enable Kyber R3 mode", default=False)
    parser.add_argument("--ml-kem-ipd", action="store_true", help="Enable ML-KEM initial public draft mode", default=False)
    parser.add_argument("--ml-kem", action="store_true", help="Enable ML-KEM final mode", default=False)
    parser.add_argument("--kats-per-mode", type=int, help="Number of KATs to generate per mode", default=25)

    return parser.parse_args()


def map_mode(file_name, mode):
    if mode == "Kyber-r3":
        # Note! See the helper shellscipt in the comment on the top of this file
        #       to generate KAT *.rsp files that match this naming scheme.
        if file_name == "PQCgenKAT_kem1024-90s":
            return "Kyber-1024-90s-r3"
        if file_name == "PQCgenKAT_kem512-90s":
            return "Kyber-512-90s-r3"
        if file_name == "PQCgenKAT_kem768-90s":
            return "Kyber-768-90s-r3"
        if file_name == "PQCgenKAT_kem1024":
            return "Kyber-1024-r3"
        if file_name == "PQCgenKAT_kem512":
            return "Kyber-512-r3"
        if file_name == "PQCgenKAT_kem768":
            return "Kyber-768-r3"
    elif mode == "ML-KEM-ipd":
        if file_name == "PQCkemKAT_3168":
            return "ML-KEM-1024-ipd"
        if file_name == "PQCkemKAT_1632":
            return "ML-KEM-512-ipd"
        if file_name == "PQCkemKAT_2400":
            return "ML-KEM-768-ipd"
    elif mode == "ML-KEM":
        if file_name == "kat_MLKEM_1024":
            return "ML-KEM-1024"
        if file_name == "kat_MLKEM_512":
            return "ML-KEM-512"
        if file_name == "kat_MLKEM_768":
            return "ML-KEM-768"
    else:
        raise Exception('Unknown mode', mode)

    raise Exception('Unknown Kyber KAT file name', file_name)


def selected_mode(args):
    modes = []

    if args.kyber_r3:
        modes.append("Kyber-r3")
    if args.ml_kem_ipd:
        modes.append("ML-KEM-ipd")
    if args.ml_kem:
        modes.append("ML-KEM")

    if len(modes) > 1:
        raise Exception("Error: More than one mode selected")

    if len(modes) == 0:
        raise Exception("Error: No mode selected")

    return modes[0]


def output_file(mode):
    if mode == "Kyber-r3":
        return "src/tests/data/pubkey/kyber_kat.vec"
    if mode == "ML-KEM-ipd":
        return "src/tests/data/pubkey/ml_kem_ipd.vec"
    if mode == "ML-KEM":
        return "src/tests/data/pubkey/ml_kem.vec"

    raise Exception("Unknown mode", mode)


def main(args = None):
    if args is None:
        return 1

    mode = selected_mode(args)

    with open(output_file(mode), 'w') as output:
        if mode == "ML-KEM":
            print("# This file was auto-generated from github.com/post-quantum-cryptography/KAT", file=output)
        else:
            print("# This file was auto-generated from the reference implemention's KATs", file=output)
        print("# See src/scripts/dev_tools/gen_kyber_kat.py\n", file=output)

        for file in args.files:
            algo_mode = map_mode(os.path.basename(os.path.splitext(file)[0]), mode)

            reader = KatReader(open(file))

            print(f"[{algo_mode}]", file=output)

            for kat in list(reader.read_kats())[:args.kats_per_mode]:
                kat = compress_kat(kat, mode)

                for key in kat.keys():
                    print(key, '=', kat[key], file=output)
                print("", file=output)

if __name__ == '__main__':
    args = parse_arguments()
    sys.exit(main(args))



#!/usr/bin/env python3

"""
Generate the Botan website

(C) 2017 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import optparse # pylint: disable=deprecated-module
import subprocess
import sys
import errno
import shutil
import tempfile
import os

def run_and_check(cmd_line, cwd=None):
    print("Executing %s ..." % (' '.join(cmd_line)))

    proc = subprocess.Popen(cmd_line,
                            cwd=cwd,
                            close_fds=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    (stdout, stderr) = proc.communicate()

    if proc.returncode != 0:
        print("Error running %s" % (' '.join(cmd_line)))
        print(stdout)
        print(stderr)
        sys.exit(1)

def rmtree_ignore_missing(path):
    try:
        shutil.rmtree(path)
    except OSError:
        # check errno?
        pass

def configure_build(botan_dir, build_dir):

    run_and_check([os.path.join(botan_dir, 'configure.py'),
                   '--with-doxygen', '--with-sphinx',
                   '--with-build-dir=%s' % (build_dir)])

def run_doxygen(tmp_dir, output_dir):
    run_and_check(['doxygen', os.path.join(tmp_dir, 'build/botan.doxy')])
    shutil.move(os.path.join(tmp_dir, 'build/docs/doxygen'), output_dir)

def run_sphinx(botan_dir, tmp_dir, output_dir):

    sphinx_config = os.path.join(botan_dir, 'src/configs/sphinx')
    sphinx_dir = os.path.join(tmp_dir, 'sphinx')
    os.mkdir(sphinx_dir)

    shutil.copyfile(os.path.join(botan_dir, 'readme.rst'),
                    os.path.join(sphinx_dir, 'index.rst'))

    for f in ['news.rst', os.path.join('doc', 'security.rst')]:
        shutil.copy(os.path.join(botan_dir, f), sphinx_dir)

    toc = """.. toctree::

   index
   news
   security
   User Guide <https://botan.randombit.net/handbook>
   API Reference <https://botan.randombit.net/doxygen>
"""

    contents_rst = open(os.path.join(sphinx_dir, 'contents.rst'), 'w', encoding='utf8')
    contents_rst.write(toc)
    contents_rst.close()

    sphinx_invoke = ['sphinx-build', '-t', 'website', '-c', sphinx_config, '-b', 'html', '-j', 'auto', '-W']

    handbook_dir = os.path.join(botan_dir, 'doc')

    run_and_check(sphinx_invoke + [sphinx_dir, output_dir])
    run_and_check(sphinx_invoke + [handbook_dir, os.path.join(output_dir, 'handbook')])

    rmtree_ignore_missing(os.path.join(output_dir, '.doctrees'))
    rmtree_ignore_missing(os.path.join(output_dir, 'handbook', '.doctrees'))
    os.remove(os.path.join(output_dir, '.buildinfo'))
    os.remove(os.path.join(output_dir, 'handbook', '.buildinfo'))

    # share _static subdirs
    shutil.rmtree(os.path.join(output_dir, 'handbook', '_static'))
    os.symlink('../_static', os.path.join(output_dir, 'handbook', '_static'))

    # Build PDF
    latex_output = os.path.join(tmp_dir, 'latex')
    run_and_check(['sphinx-build', '-c', sphinx_config, '-b', 'latex', handbook_dir, latex_output])

    # Have to run twice because TeX
    run_and_check(['pdflatex', 'botan.tex'], cwd=latex_output)
    run_and_check(['pdflatex', 'botan.tex'], cwd=latex_output)

    shutil.copy(os.path.join(latex_output, 'botan.pdf'),
                os.path.join(output_dir, 'handbook'))


def main(args):
    parser = optparse.OptionParser()

    parser.add_option('-o', '--output-dir', default=None,
                      help="Where to write output")

    (options, args) = parser.parse_args(args)

    output_dir = options.output_dir
    tmp_dir = tempfile.mkdtemp(prefix='botan_website_')

    # assumes we live in src/scripts
    botan_dir = os.path.normpath(os.path.join(os.path.dirname(__file__),
                                              "..", ".."))

    if os.access(os.path.join(botan_dir, 'configure.py'), os.X_OK) is False:
        print("Can't find configure.py in %s", botan_dir)
        return 1

    if output_dir is None:
        cwd = os.getcwd()

        if os.path.basename(cwd) == 'botan-website':
            output_dir = '.'
        else:
            output_dir = os.path.join(cwd, 'botan-website')

            try:
                os.mkdir(output_dir)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    pass
                else:
                    raise e

    for subdir in ['_static', '_sources', 'doxygen', 'handbook']:
        try:
            shutil.rmtree(os.path.join(output_dir, subdir))
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                print("Error removing dir", e)
                return 1

    configure_build(botan_dir, tmp_dir)
    run_doxygen(tmp_dir, output_dir)
    run_sphinx(botan_dir, tmp_dir, output_dir)

    for f in ['doc/pgpkey.txt', 'license.txt']:
        shutil.copy(os.path.join(botan_dir, f), output_dir)

    favicon = open(os.path.join(output_dir, 'favicon.ico'), 'w', encoding='utf8')
    # Create an empty favicon.ico file so it gets cached by browsers
    favicon.close()

    shutil.rmtree(tmp_dir)

    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/env python3

# (C) 2019 Jack Lloyd
# Botan is released under the Simplified BSD License (see license.txt)

import argparse
import subprocess
import logging
import sys
import os
import time

def script_is_disabled(script_name):
    if script_name.find('tls13') >= 0:
        return True
    if script_name.find('sslv2') >= 0:
        return True

    disabled = {
        'test-SSLv3-padding.py',
        'test-serverhello-random.py', # assumes support for SSLv2 hello
    }

    if script_name in disabled:
        return True

    slow = {
        'test-bleichenbacher-workaround.py',
        'test-client-compatibility.py',
        'test-dhe-key-share-random.py',
        'test-dhe-no-shared-secret-padding.py',
        'test-ecdhe-padded-shared-secret.py',
        'test-ecdhe-rsa-key-share-random.py',
        'test-fuzzed-plaintext.py',
        'test-invalid-client-hello-w-record-overflow.py',
        'test-invalid-client-hello.py',
        'test-large-hello.py',
    }
    if script_name in slow:
        return True

    return False

def main(args = None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser()

    # TODO generate key and spawn the server on some random port in tmp dir
    # TODO support running tls_server binary under valgrind

    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('tls-fuzzer-dir')

    args = vars(parser.parse_args(args))

    tlsfuzzer_dir = args['tls-fuzzer-dir']

    if not os.access(tlsfuzzer_dir, os.X_OK):
        raise Exception("Unable to read TLS fuzzer dir")

    tls_scripts_dir = os.path.join(tlsfuzzer_dir, 'scripts')
    if not os.access(tlsfuzzer_dir, os.X_OK):
        raise Exception("Unable to read TLS fuzzer scripts dir")

    scripts = sorted(os.listdir(tls_scripts_dir))

    procs = {}

    for script in scripts:
        if script_is_disabled(script):
            logging.debug('Skipping %s' % (script))
            continue

        procs[script] = subprocess.Popen([sys.executable, os.path.join(tls_scripts_dir, script)],
                                         cwd=tlsfuzzer_dir,
                                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    results = {}

    while len(results) != len(procs):
        time.sleep(.5)
        for (script, proc) in procs.items():

            if script in results:
                continue

            if proc.poll() != None:
                rv = proc.returncode
                results[script] = rv
                if rv == 0:
                    print("PASS %s" % (script))
                else:
                    print("FAIL %s" % (script))
                sys.stdout.flush()
    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

# Converts the json files of ACVP KATs to a format that can be used by Botan tests.
# This script relies on unlicensed code from Markku-Juhani O. Saarinen
# taken from https://github.com/mjosaarinen/py-acvp-pqc/blob/main/test_mlkem.py
#
# (C) 2024 Jack Lloyd
# (C) 2024 Amos Treiber - Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

import binascii
import hashlib
import json
import argparse
import sys

def shake_256_16(v):
    # v is assumed to be hex
    h = hashlib.shake_256()
    h.update(binascii.unhexlify(v))
    return h.hexdigest(16)

def mlkem_load_keygen(req_fn, res_fn):
    with open(req_fn) as f:
        keygen_req = json.load(f)
    with open(res_fn) as f:
        keygen_res = json.load(f)

    keygen_kat = []
    for qtg in keygen_req['testGroups']:
        alg = qtg['parameterSet']
        tgid = qtg['tgId']

        rtg = None
        for tg in keygen_res['testGroups']:
            if tg['tgId'] == tgid:
                rtg = tg['tests']
                break

        for qt in qtg['tests']:
            tcid = qt['tcId']
            for t in rtg:
                if t['tcId'] == tcid:
                    qt.update(t)
            qt['parameterSet'] = alg
            keygen_kat += [qt]

    return keygen_kat

def mlkem_load_encdec(req_fn, res_fn):
    with open(req_fn) as f:
        encdec_req = json.load(f)
    with open(res_fn) as f:
        encdec_res = json.load(f)

    encaps_kat = []
    decaps_kat = []
    for qtg in encdec_req['testGroups']:
        alg = qtg['parameterSet']
        func = qtg['function']
        tgid = qtg['tgId']

        rtg = None
        for tg in encdec_res['testGroups']:
            if tg['tgId'] == tgid:
                rtg = tg['tests']
                break

        for qt in qtg['tests']:
            tcid = qt['tcId']
            for t in rtg:
                if t['tcId'] == tcid:
                    qt.update(t)
            qt['parameterSet'] = alg
            if func == 'encapsulation':
                encaps_kat += [qt]
            elif func == 'decapsulation':
                qt['dk'] = qtg['dk']
                decaps_kat += [qt]
            else:
                print('ERROR: Unkonwn function:', func)

    return (encaps_kat, decaps_kat)

def group_by_parameter_set(keygen_kat):
    grouped_kat = {}
    for kat in keygen_kat:
        parameter_set = kat['parameterSet']
        if parameter_set not in grouped_kat:
            grouped_kat[parameter_set] = []
        grouped_kat[parameter_set].append(kat)
    return grouped_kat

def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--keygen-directory", type=str, help="Directory 'ML-KEM-keyGen-FIPS203' containing the JSON files")
    parser.add_argument("--encapdecap-directory", type=str, help="Directory 'ML-KEM-encapDecap-FIPS203' containing the JSON files")

    return parser.parse_args()

def compress_kat_keygen(kat):
    kat['Z']  = kat.pop('z')
    kat['D']  = kat.pop('d')
    kat['EK'] = shake_256_16(kat.pop('ek'))
    kat['DK'] = shake_256_16(kat['D'] + kat['Z'])
    del kat['dk'] # remove the original decaps key (Botan uses the private seed as storage format)

    return kat

def compress_kat_encaps(kat):
    kat['EK'] = kat.pop('ek')
    kat['M']  = kat.pop('m')
    kat['K']  = kat.pop('k')
    kat['C']  = shake_256_16(kat.pop('c'))

    return kat

def main(args = None):
    if args is None:
        return 1

    if args.keygen_directory is not None:

        keygen_kat = mlkem_load_keygen(
                    args.keygen_directory + '/prompt.json',
                    args.keygen_directory + '/expectedResults.json')
        with open("src/tests/data/pubkey/ml_kem_acvp_keygen.vec", 'w') as output:
            print("# This file was auto-generated from the ACVP KATs", file=output)
            print("# See src/scripts/dev_tools/gen_mlkem_acvp_kat.py\n", file=output)

            for paramset, kat_by_paramset in group_by_parameter_set(keygen_kat).items():
                print(f"[{paramset}]", file=output)

                for kat in kat_by_paramset:
                    kat = compress_kat_keygen(kat)
                    for key in kat.keys():
                        if key in ["Z", "D", "EK", "DK"]:
                            print(key, '=', kat[key], file=output)
                    print("", file=output)
    if args.encapdecap_directory is not None:
        encaps_kat, decaps_kat = mlkem_load_encdec(
                    args.encapdecap_directory + '/prompt.json',
                    args.encapdecap_directory + '/expectedResults.json')

        with open("src/tests/data/pubkey/ml_kem_acvp_encap.vec", 'w') as output:
            print("# This file was auto-generated from the ACVP KATs", file=output)
            print("# See src/scripts/dev_tools/gen_mlkem_acvp_kat.py\n", file=output)

            for paramset, kat_by_paramset in group_by_parameter_set(encaps_kat).items():
                print(f"[{paramset}]", file=output)

                for kat in kat_by_paramset:
                    kat = compress_kat_encaps(kat)
                    for key in kat.keys():
                        if key in ["M", "EK", "K", "C"]:
                            print(key, '=', kat[key], file=output)
                    print("", file=output)

        with open("src/tests/data/pubkey/ml_kem_acvp_decap.vec", 'w') as output:
            print("# This file was auto-generated from the ACVP KATs", file=output)
            print("# See src/scripts/dev_tools/gen_mlkem_acvp_kat.py\n", file=output)

            # We cannot use this KAT at the moment as it does not provide the
            # private seeds for the decapsulation operation.

            # for paramset, kat_by_paramset in group_by_parameter_set(decaps_kat).items():
            #     print(f"[{paramset}]", file=output)

            #     for kat in kat_by_paramset:
            #         # No compressions possible in decaps
            #         for key in ["C", "DK", "K"]:
            #             if key.lower() in kat.keys():
            #                 print(key, '=', kat[key.lower()], file=output)
            #         print("", file=output)



if __name__ == '__main__':
    args = parse_arguments()
    sys.exit(main(args))



#!/usr/bin/env python3

"""
Unittests for Botan Python scripts.

Requires Python 3.

(C) 2017 Simon Warta (Kullo GmbH)

Botan is released under the Simplified BSD License (see license.txt)
"""

import sys
import unittest

sys.path.append("../..") # Botan repo root
from configure import AmalgamationHelper # pylint: disable=wrong-import-position
from configure import ModulesChooser # pylint: disable=wrong-import-position

class AmalgamationHelperTests(unittest.TestCase):
    def test_matcher_std_includes(self):
        self.assertEqual(AmalgamationHelper.is_unconditional_std_include("#include <string>"), "string")
        self.assertEqual(AmalgamationHelper.is_unconditional_std_include("#include <string> // comment"), "string")

        self.assertEqual(AmalgamationHelper.is_unconditional_std_include("#include <myfile.h>"), None)
        self.assertEqual(AmalgamationHelper.is_unconditional_std_include("#include <unistd.h>"), None)
        self.assertEqual(AmalgamationHelper.is_unconditional_std_include("  #include <string>"), None)

    def test_matcher_botan_include(self):
        self.assertEqual(AmalgamationHelper.is_botan_include("#include <botan/oids.h>"),
                         "oids.h")
        self.assertEqual(AmalgamationHelper.is_botan_include("#include <botan/internal/socket.h>"),
                         "internal/socket.h")
        self.assertEqual(AmalgamationHelper.is_botan_include("#include <botan/oids.h> // comment"),
                         "oids.h")
        self.assertEqual(AmalgamationHelper.is_botan_include("#include <botan/internal/socket.h> // comment"),
                         "internal/socket.h")
        self.assertEqual(AmalgamationHelper.is_botan_include("  #include <botan/oids.h>"),
                         "oids.h")
        self.assertEqual(AmalgamationHelper.is_botan_include("  #include <botan/internal/socket.h>"),
                         "internal/socket.h")

        self.assertEqual(AmalgamationHelper.is_botan_include("#include <string>"), None)
        self.assertEqual(AmalgamationHelper.is_botan_include("#include <myfile.h>"), None)
        self.assertEqual(AmalgamationHelper.is_botan_include("#include <unistd.h>"), None)

    def test_matcher_any_includes(self):
        self.assertEqual(AmalgamationHelper.is_any_include("#include <string>"), "string")
        self.assertEqual(AmalgamationHelper.is_any_include("#include <myfile.h>"), "myfile.h")
        self.assertEqual(AmalgamationHelper.is_any_include("#include <unistd.h>"), "unistd.h")
        self.assertEqual(AmalgamationHelper.is_any_include("#include <botan/oids.h>"),
                         "botan/oids.h")
        self.assertEqual(AmalgamationHelper.is_any_include("  #include <string>"), "string")
        self.assertEqual(AmalgamationHelper.is_any_include("  #include <myfile.h>"), "myfile.h")
        self.assertEqual(AmalgamationHelper.is_any_include("  #include <unistd.h>"), "unistd.h")
        self.assertEqual(AmalgamationHelper.is_any_include("  #include <botan/oids.h>"),
                         "botan/oids.h")
        self.assertEqual(AmalgamationHelper.is_any_include("#include <string> // comment"), "string")
        self.assertEqual(AmalgamationHelper.is_any_include("#include <myfile.h> // comment"), "myfile.h")
        self.assertEqual(AmalgamationHelper.is_any_include("#include <unistd.h> // comment"), "unistd.h")
        self.assertEqual(AmalgamationHelper.is_any_include("#include <botan/oids.h> // comment"),
                         "botan/oids.h")

class ModulesChooserResolveDependencies(unittest.TestCase):
    def test_base(self):
        available_modules = set(["A", "B"])
        table = {
            "A": [],
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A"]))

    def test_no_dependencies_defined(self):
        available_modules = set(["A", "B"])
        table = {
            "A": [],
        }
        with self.assertRaises(KeyError):
            ModulesChooser.resolve_dependencies(available_modules, table, "B")

        available_modules = set(["A", "B"])
        table = {
            "A": ["B"],
        }
        with self.assertRaises(KeyError):
            ModulesChooser.resolve_dependencies(available_modules, table, "A")

    def test_add_dependency(self):
        available_modules = set(["A", "B"])
        table = {
            "A": ["B"],
            "B": []
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A", "B"]))

    def test_add_dependencies_two_levels(self):
        available_modules = set(["A", "B", "C"])
        table = {
            "A": ["B"],
            "B": ["C"],
            "C": []
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A", "B", "C"]))

    def test_circular(self):
        available_modules = set(["A", "B", "C"])
        table = {
            "A": ["B"],
            "B": ["C"],
            "C": ["A"]
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A", "B", "C"]))

    def test_not_available(self):
        available_modules = set(["A", "C"])
        table = {
            "A": ["B"],
            "B": ["C"],
            "C": ["A"]
        }
        ok, _ = ModulesChooser.resolve_dependencies(available_modules, table, "B")
        self.assertFalse(ok)

    def test_dependency_not_available(self):
        available_modules = set(["A", "C"])
        table = {
            "A": ["B"],
            "B": ["C"],
            "C": ["A"]
        }
        ok, _ = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertFalse(ok)

    def test_dependency2_not_available(self):
        available_modules = set(["A", "B"])
        table = {
            "A": ["B"],
            "B": ["C"],
            "C": ["A"]
        }
        ok, _ = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertFalse(ok)

    def test_dependency_choices(self):
        available_modules = set(["A", "B", "C"])
        table = {
            "A": ["B|C"],
            "B": [],
            "C": []
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertTrue(modules == set(["A", "B"]) or modules == set(["A", "C"]))

    def test_dependency_prefer_existing(self):
        available_modules = set(["A", "B", "C"])
        table = {
            "A": ["C", "B|C"],
            "B": [],
            "C": []
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A", "C"]))

    def test_dependency_prefer_existing2(self):
        available_modules = set(["A", "B", "C"])
        table = {
            "A": ["B", "B|C"],
            "B": [],
            "C": []
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A", "B"]))

    def test_dependency_choices_impossible(self):
        available_modules = set(["A", "C"])
        table = {
            "A": ["B|C"],
            "B": [],
            "C": []
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A", "C"]))

    def test_dependency_choices_impossible2(self):
        available_modules = set(["A", "B"])
        table = {
            "A": ["B|C"],
            "B": [],
            "C": []
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "A")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["A", "B"]))

    def test_deep(self):
        available_modules = set(["A", "B", "C", "E", "G"])
        table = {
            "A": ["B|C"],
            "B": ["D"],
            "C": ["E"],
            "D": [],
            "E": ["F|G"],
            "F": ["A", "B"],
            "G": ["A", "G"]
        }
        ok, modules = ModulesChooser.resolve_dependencies(available_modules, table, "G")
        self.assertTrue(ok)
        self.assertEqual(modules, set(["G", "A", "C", "E"]))


if __name__ == '__main__':
    unittest.TestCase.longMessage = True
    unittest.main()



# Script to run inside the CI container to test the botan
# TLS client/server with TLS-Anvil
#
# (C) 2023 Jack Lloyd
# (C) 2023 Fabian Albert, Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
import sys
import argparse
import os
import subprocess


class Config:
    """ Hardcoded configurations for this CI script """
    tls_anvil_docker_image = "ghcr.io/tls-attacker/tlsanvil"
    tls_anvil_version_tag = "@sha256:e9abe034e6b1dac7fe204d524db338f379087a16aca71e94dc7b51ac835bb53f" # v1.2.2 + HelloRetry test fix
    key_and_cert_storage_path = "/tmp/"
    test_suite_results_dest = "."
    test_suite_results_dir_name = "TestSuiteResults"
    tmp_key_file_name = "tmp_rsa_key.pem"
    tmp_cert_file_name = "tmp_rsa_cert.pem"
    server_dest_ip = "127.0.0.1"
    server_dest_port = 4433
    botan_server_log = "./logs/botan_server.log"
    botan_config_args = ["--compiler-cache=ccache", "--build-targets=static,cli",
                          "--without-documentation", "--with-boost"]

def group_output(group_title: str, func):
    """
    Wraps a function to be called within a GitHub actions group, so that
    the console output is expandable.

    Returns the wrapped function
    """
    def wrapped_func(*args, **kwargs):
        print(f"::group::{group_title}", flush=True)
        ret = func(*args, **kwargs)
        print("\n::endgroup::", flush=True)
        return ret
    return wrapped_func


def create_cert_and_key(botan_dir_path):
    """
    Create a X.509 certificate and associated RSA key at Config.key_and_cert_storage_path
    using Botan's CLI.

    Returns: (<cert path>, <key path>)
    """

    key_path = os.path.join(Config.key_and_cert_storage_path, Config.tmp_key_file_name)
    cert_path = os.path.join(Config.key_and_cert_storage_path, Config.tmp_cert_file_name)

    with open(key_path, 'w', encoding='utf-8') as keyfile:
        subprocess.run([botan_dir_path, "keygen", "--algo=RSA", "--params=2048"], stdout=keyfile, check=True)

    with open(cert_path, 'w', encoding='utf-8') as certfile:
        subprocess.run([botan_dir_path, "gen_self_signed", key_path, "localhost"], stdout=certfile, check=True)

    return (cert_path, key_path)


def build_botan(botan_dir: str, parallel_jobs: int) -> str:
    """
    Configure and build botan.

    Returns the botan executable path
    """
    group_output("Configure Botan", subprocess.run)(["python3", "./configure.py"] + Config.botan_config_args, check=True, cwd=botan_dir)
    group_output("Build Botan with Make", subprocess.run)(["make", f"-j{parallel_jobs}"], check=True, cwd=botan_dir)

    return os.path.join(botan_dir, "botan")


def server_test(botan_dir_path: str, parallel: int):
    """ Test the Botan TLS server """
    cert_path, key_path = create_cert_and_key(botan_dir_path)
    docker_img = f"{Config.tls_anvil_docker_image}{Config.tls_anvil_version_tag}"

    group_output("Pull TLS-Anvil image", subprocess.run)(["docker", "pull", docker_img], check=True)

    tls_anvil_cmd = [
        "docker", "run",
        "--network", "host",
        "-v", f"{Config.test_suite_results_dest}:/output",
        docker_img,
        "-strength", "1",
        "-parallelHandshakes", str(parallel),
        "-disableTcpDump",
        "-outputFolder", os.path.join(Config.test_suite_results_dest, Config.test_suite_results_dir_name),
        "-connectionTimeout", "5000",
        "server", "-connect", f"{Config.server_dest_ip}:{Config.server_dest_port}"
    ]

    botan_server_cmd = [
        botan_dir_path, "tls_http_server", cert_path, key_path, f"--port={Config.server_dest_port}"
    ]

    os.makedirs(os.path.dirname(Config.botan_server_log), exist_ok=True)

    # Run Botan and test is with TLS-Anvil
    with open(Config.botan_server_log, 'w', encoding='utf-8') as server_log_file:
        botan_server_process = subprocess.Popen(botan_server_cmd, stdout=server_log_file, stderr=server_log_file)
        subprocess.run(tls_anvil_cmd, check=True)
        botan_server_process.kill()


def client_test(botan_dir_path: str, parallel: int):
    """ Test the Botan TLS server """
    raise NotImplementedError("Client tests not yet implemented")


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("--botan-dir", help="Botan base directory", required=True)
    parser.add_argument("--test-target", help="The TLS side to test", choices=['client', 'server'], required=True)
    parser.add_argument("--parallel", help="The number of parallel handshakes", type=int, default=1)

    args = vars(parser.parse_args(args))

    if not os.path.isdir(args["botan_dir"]):
        raise FileNotFoundError(f"Unable to find '{args['botan_dir']}'")

    botan_exe_path = build_botan(args["botan_dir"], args["parallel"])

    if args["test_target"] == "server":
        server_test(botan_exe_path, args["parallel"])
    elif args["test_target"] == "client":
        client_test(botan_exe_path, args["parallel"])

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

# (C) 2017,2018 Jack Lloyd
# Botan is released under the Simplified BSD License (see license.txt)

import sys
import os
import subprocess
import optparse # pylint: disable=deprecated-module
import stat
import multiprocessing
import time
import tempfile

def run_fuzzer_gdb(args):
    (fuzzer_bin, corpus_file) = args

    gdb_proc = subprocess.Popen(['gdb', '--quiet', '--return-child-result', fuzzer_bin],
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                close_fds=True)

    gdb_commands = ('run < %s\nbt\nquit\n' % (corpus_file)).encode('ascii')

    (stdout, stderr) = gdb_proc.communicate(gdb_commands)

    if gdb_proc.returncode == 0:
        return (0, '', '')

    return (corpus_file, gdb_proc.returncode, stdout.decode('ascii'), stderr.decode('ascii'))

def run_fuzzer(args):
    (fuzzer_bin, corpus_file) = args
    corpus_fd = open(corpus_file, 'rb')
    fuzzer_proc = subprocess.Popen([fuzzer_bin], stdin=corpus_fd,
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    (stdout, stderr) = fuzzer_proc.communicate()
    corpus_fd.close()
    return (corpus_file, fuzzer_proc.returncode, stdout.decode('ascii'), stderr.decode('ascii'))

def run_fuzzer_many_files(fuzzer_bin, corpus_files):
    fuzzer_proc = subprocess.Popen([fuzzer_bin] + corpus_files, stdin=None,
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    (stdout, stderr) = fuzzer_proc.communicate()
    return (fuzzer_proc.returncode, stdout.decode('ascii'), stderr.decode('ascii'))

def main(args=None):
    if args is None:
        args = sys.argv

    parser = optparse.OptionParser(
        usage='Usage: %prog [options] corpus_dir fuzzers_dir',
    )

    parser.add_option('--gdb', action='store_true',
                      help='Run under GDB and capture backtraces')

    parser.add_option('--one-at-a-time', action='store_true', default=False,
                      help='Test one corpus input at a time')

    (options, args) = parser.parse_args(args)

    if len(args) != 3:
        parser.print_usage()
        return 1

    if options.gdb and not options.one_at_a_time:
        print("Option --gdb requires --one-at-a-time")
        return 1

    corpus_dir = args[1]
    fuzzer_dir = args[2]

    if not os.access(corpus_dir, os.R_OK):
        print("Error could not access corpus directory '%s'" % (corpus_dir))
        return 1

    if not os.access(fuzzer_dir, os.R_OK):
        print("Error could not access fuzzers directory '%s'" % (fuzzer_dir))
        return 1

    fuzzers = set([])
    for fuzzer in os.listdir(fuzzer_dir):
        if fuzzer.endswith('.zip'):
            continue
        fuzzers.add(fuzzer)

    corpii = set([])
    for corpus in os.listdir(corpus_dir):
        # Ignore regular files in toplevel dir
        if not stat.S_ISDIR(os.stat(os.path.join(corpus_dir, corpus)).st_mode):
            continue

        if corpus == '.git':
            continue

        corpii.add(corpus)

    fuzzers_without_corpus = fuzzers - corpii
    corpus_without_fuzzers = corpii - fuzzers

    for f in sorted(list(fuzzers_without_corpus)):
        print("Warning: Fuzzer %s has no corpus" % (f))
    for c in sorted(list(corpus_without_fuzzers)):
        print("Warning: Corpus %s has no fuzzer" % (c))

    fuzzers_with_corpus = fuzzers & corpii

    crash_count = 0
    stderr_count = 0
    stdout_count = 0

    if options.one_at_a_time:
        pool = multiprocessing.Pool(multiprocessing.cpu_count() * 2)
        chunk_size = 32 # arbitrary

        run_fuzzer_func = run_fuzzer_gdb if options.gdb else run_fuzzer

        for fuzzer in sorted(list(fuzzers_with_corpus)):
            fuzzer_bin = os.path.join(fuzzer_dir, fuzzer)
            corpus_subdir = os.path.join(corpus_dir, fuzzer)
            corpus_files = [os.path.join(corpus_subdir, l) for l in sorted(list(os.listdir(corpus_subdir)))]

            # We have to do this hack because multiprocessing's Pool.map doesn't support
            # passing any initial arguments, just the single iteratable
            map_args = [(fuzzer_bin, f) for f in corpus_files]

            start = time.time()

            for result in pool.map(run_fuzzer_func, map_args, chunk_size):
                (corpus_file, retcode, stdout, stderr) = result

                if retcode != 0:
                    print("Fuzzer %s crashed with input %s returncode %d" % (fuzzer, corpus_file, retcode))
                    crash_count += 1

                if stdout:
                    print("Fuzzer %s produced stdout on input %s:\n%s" % (fuzzer, corpus_file, stdout))
                    stdout_count += 1

                if stderr:
                    print("Fuzzer %s produced stderr on input %s:\n%s" % (fuzzer, corpus_file, stderr))
                    stderr_count += 1

            duration = time.time() - start
            print("Tested fuzzer %s with %d test cases, %d crashes in %.02f seconds" % (
                fuzzer, len(corpus_files), crash_count, duration))
            crash_count = 0
            sys.stdout.flush()
    else:
        # Generate a random corpus for fuzzers without a corpus
        random_corpus_dir = tempfile.mkdtemp(prefix='fuzzer_corpus_')

        slow_fuzzers = ['invert', 'ecc_p521', 'pow_mod']
        random_corpus_size = 1000
        random_corpus_size_for_slow_fuzzers = 100

        for i in range(random_corpus_size):
            random_input = os.urandom(i)
            fd = open(os.path.join(random_corpus_dir, 'input_%d' % (i)), 'wb')
            fd.write(random_input)
            fd.close()

        for fuzzer in sorted(list(fuzzers)):
            fuzzer_bin = os.path.join(fuzzer_dir, fuzzer)

            if fuzzer in fuzzers_with_corpus:
                corpus_subdir = os.path.join(corpus_dir, fuzzer)
            else:
                corpus_subdir = random_corpus_dir

            corpus_files = [os.path.join(corpus_subdir, l) for l in sorted(list(os.listdir(corpus_subdir)))]

            if fuzzer in slow_fuzzers:
                corpus_files = corpus_files[:random_corpus_size_for_slow_fuzzers]

            start = time.time()

            (retcode, stdout, stderr) = run_fuzzer_many_files(fuzzer_bin, corpus_files)

            if retcode != 0:
                print("Fuzzer %s crashed returncode %d" % (fuzzer, retcode))
                crash_count += 1

            if stdout:
                print("Fuzzer %s produced stdout:\n%s" % (fuzzer, stdout))
                stdout_count += 1

            if stderr:
                print("Fuzzer %s produced stderr:\n%s" % (fuzzer, stderr))
                stderr_count += 1

            duration = time.time() - start

            print("Tested fuzzer %s with %d test cases, %d crashes in %.02f seconds" % (
                fuzzer, len(corpus_files), crash_count, duration))
            crash_count = 0

    if crash_count > 0 or stderr_count > 0 or stdout_count > 0:
        print("Ran fuzzer tests, %d crashes %d stdout %d stderr" % (crash_count, stdout_count, stderr_count))
        return 2
    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2016 Jack Lloyd
(C) 2017 Fabian Weissberg, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import sys
import datetime
import re
from collections import defaultdict

def format_oid(oid):
    #return '"' + oid + '"'
    return "{" + oid.replace('.', ', ') + '}'

def format_map(m, for_oid = False):
    s = ''
    for k in sorted(m.keys()):
        v = m[k]

        if len(s) > 0:
            s += '      '

        if for_oid:
            s += '{"%s", OID(%s)},\n' % (k,format_oid(v))
        else:
            s += '{OID(%s), "%s"},\n' % (format_oid(k),v)

    s = s[:-2] # chomp last two chars

    return s


def format_as_map(oid2str, str2oid):
   return """/*
* OID maps
*
* This file was automatically generated by %s on %s
*
* All manual edits to this file will be lost. Edit the script
* then regenerate this source file.
*
* Botan is released under the Simplified BSD License (see license.txt)
*/

#include <botan/internal/oid_map.h>
#include <unordered_map>

namespace Botan {

std::unordered_map<OID, std::string> OID_Map::load_oid2str_map() {
   return std::unordered_map<OID, std::string>{

      %s};
}

std::unordered_map<std::string, OID> OID_Map::load_str2oid_map() {
   return std::unordered_map<std::string, OID>{

      %s};
}

}  // namespace Botan""" % (
    sys.argv[0],
    datetime.date.today().strftime("%Y-%m-%d"),
    format_map(oid2str),
    format_map(str2oid, True))

def format_dn_ub_map(dn_ub, oid2str):
    s = ''
    for k in sorted(dn_ub.keys()):
        v = dn_ub[k]

        expr = "   {OID({%s}), %s}, " % (k.replace('.', ', '), v)
        s += expr
        s += ' '*(32 - len(expr))
        s += ' // %s\n' % (oid2str[k])

    # delete last ',' and \n
    idx = s.rfind(',')
    if idx != -1:
        s = s[:idx] + ' ' + s[idx+1:-1]

    return s


def format_dn_ub_as_map(dn_ub, oid2str):
    return """/*
* DN_UB maps: Upper bounds on the length of DN strings
*
* This file was automatically generated by %s on %s
*
* All manual edits to this file will be lost. Edit the script
* then regenerate this source file.
*
* Botan is released under the Simplified BSD License (see license.txt)
*/

#include <botan/pkix_types.h>

#include <botan/asn1_obj.h>
#include <map>

namespace Botan {

namespace {

/**
 * Upper bounds for the length of distinguished name fields as given in RFC 5280, Appendix A.
 * Only OIDS recognized by botan are considered, so far.
 * Maps OID string representations instead of human readable strings in order
 * to avoid an additional lookup.
 */
const std::map<OID, size_t> DN_UB = {
%s
};

}  // namespace

//static
size_t X509_DN::lookup_ub(const OID& oid) {
   auto ub_entry = DN_UB.find(oid);
   if(ub_entry != DN_UB.end()) {
      return ub_entry->second;
   } else {
      return 0;
   }
}

}  // namespace Botan""" % (sys.argv[0], datetime.date.today().strftime("%Y-%m-%d"),
       format_dn_ub_map(dn_ub,oid2str))


def format_set_map(m):
    s = ''
    for k in sorted(m.keys()):
        v = m[k]

        if len(s) > 0:
            s += '   '

        s += '{ "%s", {' % k
        for pad in v:
            s += '"%s", ' % pad
        if len(v) != 0:
            s = s[:-2]
        s += '} },\n'
    s = s[:-1]
    return s


def main(args = None):
    """ Print header files (oids.cpp, dn_ub.cpp) depending on the first argument and on src/build-data/oids.txt

        Choose 'oids' to print oids.cpp, needs to be written to src/lib/asn1/oids.cpp
        Choose 'dn_ub' to print dn_ub.cpp, needs to be written to src/lib/x509/X509_dn_ub.cpp
    """
    if args is None:
        args = sys.argv
    if len(args) < 2:
        print("Use either 'oids' or 'dn_ub' as first argument")
        return 1

    oid_lines = open('./src/build-data/oids.txt').readlines()

    oid_re = re.compile(r"^([0-9][0-9.]+) += +([A-Za-z0-9_\./\(\), -]+)(?: = )?([0-9]+)?$")
    hdr_re = re.compile(r"^\[([a-z0-9_]+)\]$")

    oid2str = {}
    str2oid = {}
    dn_ub = {}
    cur_hdr = None

    for line in oid_lines:
        line = line.strip()
        if len(line) == 0:
            continue

        if line[0] == '#':
            continue

        match = hdr_re.match(line)
        if match is not None:
            cur_hdr = match.group(1)
            continue

        match = oid_re.match(line)
        if match is None:
            raise Exception(line)

        oid = match.group(1)
        nam = match.group(2)

        if oid in str2oid:
            print("Duplicated OID", oid, name, oid2str[oid])
            sys.exit() # hard error
        else:
            oid2str[oid] = nam

        # parse upper bounds for DNs
        if cur_hdr == "dn":
            if match.lastindex < 3:
                raise Exception("Could not find an upper bound for DN " + match.group(1))
            dn_ub[oid] = match.group(3)

        if nam in str2oid:
            #str2oid[nam] = oid
            pass
        else:
            str2oid[nam] = oid

    if args[1] == "oids":
        print(format_as_map(oid2str, str2oid))
    elif args[1] == "dn_ub":
        print(format_dn_ub_as_map(dn_ub,oid2str))
    else:
        print("Unknown command: try oids or dn_ub")

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
Botan install script

(C) 2014,2015,2017 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import errno
import json
import logging
import optparse # pylint: disable=deprecated-module
import os
import shutil
import sys
import traceback

def parse_command_line(args):

    parser = optparse.OptionParser()

    parser.add_option('--verbose', action='store_true', default=False,
                      help='Show debug messages')
    parser.add_option('--quiet', action='store_true', default=False,
                      help='Show only warnings and errors')

    build_group = optparse.OptionGroup(parser, 'Source options')
    build_group.add_option('--build-dir', metavar='DIR', default='build',
                           help='Location of build output (default \'%default\')')
    parser.add_option_group(build_group)

    install_group = optparse.OptionGroup(parser, 'Install options')
    install_group.add_option('--umask', metavar='MASK', default='022',
                             help='Umask to set (default %default)')
    parser.add_option_group(install_group)

    (options, args) = parser.parse_args(args)

    def log_level():
        if options.verbose:
            return logging.DEBUG
        if options.quiet:
            return logging.WARNING
        return logging.INFO

    logging.getLogger().setLevel(log_level())

    return (options, args)


class PrependDestdirError(Exception):
    pass


def is_subdir(path, subpath):
    return os.path.relpath(path, start=subpath).startswith("..")


def prepend_destdir(path):
    """
    Needed because os.path.join() discards the first path if the
    second one is absolute, which is usually the case here. Still, we
    want relative paths to work and leverage the os awareness of
    os.path.join().
    """
    destdir = os.environ.get('DESTDIR', "")

    if destdir:
        # DESTDIR is non-empty, but we only join absolute paths on UNIX-like file systems
        if os.path.sep != "/":
            raise PrependDestdirError("Only UNIX-like file systems using forward slash " \
                                      "separator supported when DESTDIR is set.")
        if not os.path.isabs(path):
            raise PrependDestdirError("--prefix must be an absolute path when DESTDIR is set.")

        path = os.path.normpath(path)
        # Remove / or \ prefixes if existent to accomodate for os.path.join()
        path = path.lstrip(os.path.sep)
        path = os.path.join(destdir, path)

        if not is_subdir(destdir, path):
            raise PrependDestdirError("path escapes DESTDIR (path='%s', destdir='%s')" % (path, destdir))

    return path


def makedirs(dirname, exist_ok=True):
    try:
        logging.debug('Creating directory %s', dirname)
        os.makedirs(dirname)
    except OSError as ex:
        if ex.errno != errno.EEXIST or not exist_ok:
            raise ex

# Clear link and create new one
def force_symlink(target, linkname):
    try:
        os.unlink(linkname)
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            raise ex
    os.symlink(target, linkname)

def calculate_exec_mode(options):
    out = 0o777
    if 'umask' in os.__dict__:
        umask = int(options.umask, 8)
        logging.debug('Setting umask to %s', oct(umask))
        os.umask(int(options.umask, 8))
        out &= (umask ^ 0o777)
    return out

def main(args):
    logging.basicConfig(stream=sys.stdout,
                        format='%(levelname) 7s: %(message)s')

    (options, args) = parse_command_line(args)

    exe_mode = calculate_exec_mode(options)
    build_dir = options.build_dir

    def copy_file(src, dst):
        logging.debug('Copying %s to %s', src, dst)
        shutil.copyfile(src, dst)

    def copy_executable(src, dst):
        copy_file(src, dst)
        logging.debug('Make %s executable', dst)
        os.chmod(dst, exe_mode)

    with open(os.path.join(build_dir, 'build_config.json'), encoding='utf8') as f:
        cfg = json.load(f)

    ver_major = int(cfg['version_major'])
    ver_minor = int(cfg['version_minor'])
    ver_patch = int(cfg['version_patch'])
    target_os = cfg['os']
    build_shared_lib = bool(cfg['build_shared_lib'])
    build_static_lib = bool(cfg['build_static_lib'])
    build_cli = bool(cfg['build_cli_exe'])
    out_dir = cfg['out_dir']

    bin_dir = cfg['bindir']
    lib_dir = cfg['libdir']
    target_include_dir = cfg['installed_include_dir']
    pkgconfig_dir = os.path.join(lib_dir, 'pkgconfig')
    cmake_dir = os.path.join(lib_dir, 'cmake', 'Botan-%s' % cfg["version"])

    prefix = cfg['prefix']

    for d in [prefix, lib_dir, bin_dir, target_include_dir]:
        makedirs(prepend_destdir(d))

    for header in cfg['public_headers']:
        full_header_path = os.path.join(cfg['public_include_path'], header)
        copy_file(full_header_path,
                  prepend_destdir(os.path.join(target_include_dir, header)))

    for header in cfg['external_headers']:
        full_header_path = os.path.join(cfg['external_include_path'], header)
        copy_file(full_header_path,
                  prepend_destdir(os.path.join(target_include_dir, header)))

    if build_static_lib:
        static_lib = cfg['static_lib_name']
        copy_file(os.path.join(out_dir, static_lib),
                  prepend_destdir(os.path.join(lib_dir, os.path.basename(static_lib))))

    if build_shared_lib:
        if target_os == "windows":
            libname = cfg['libname']
            soname_base = libname + '.dll'
            implib = cfg['implib_name']
            copy_executable(os.path.join(out_dir, soname_base),
                            prepend_destdir(os.path.join(bin_dir, soname_base)))
            copy_file(os.path.join(out_dir, implib),
                      prepend_destdir(os.path.join(lib_dir, os.path.basename(implib))))
        elif target_os == "mingw":
            shared_lib_name = cfg['shared_lib_name']
            copy_executable(os.path.join(out_dir, shared_lib_name),
                            prepend_destdir(os.path.join(bin_dir, shared_lib_name)))
            implib_name = shared_lib_name + '.a'
            copy_executable(os.path.join(out_dir, implib_name),
                            prepend_destdir(os.path.join(lib_dir, implib_name)))
        else:
            soname_patch = cfg['soname_patch']
            soname_abi = cfg['soname_abi']
            soname_base = cfg['soname_base']

            copy_executable(os.path.join(out_dir, soname_patch),
                            prepend_destdir(os.path.join(lib_dir, soname_patch)))

            if cfg['symlink_shared_lib']:
                prev_cwd = os.getcwd()
                try:
                    os.chdir(prepend_destdir(lib_dir))
                    force_symlink(soname_patch, soname_abi)
                    force_symlink(soname_patch, soname_base)
                finally:
                    os.chdir(prev_cwd)

    if build_cli:
        copy_executable(cfg['cli_exe'], prepend_destdir(os.path.join(bin_dir, cfg['cli_exe_name'])))

    if 'botan_pkgconfig' in cfg:
        makedirs(prepend_destdir(pkgconfig_dir))
        copy_file(cfg['botan_pkgconfig'],
                  prepend_destdir(os.path.join(pkgconfig_dir, os.path.basename(cfg['botan_pkgconfig']))))

    if 'botan_cmake_config' in cfg and 'botan_cmake_version_config' in cfg:
        makedirs(prepend_destdir(cmake_dir))
        copy_file(cfg['botan_cmake_config'],
                  prepend_destdir(os.path.join(cmake_dir, os.path.basename(cfg['botan_cmake_config']))))
        copy_file(cfg['botan_cmake_version_config'],
                  prepend_destdir(os.path.join(cmake_dir, os.path.basename(cfg['botan_cmake_version_config']))))

    if 'ffi' in cfg['mod_list'] and cfg['build_shared_lib'] is True and cfg['install_python_module'] is True:
        for ver in cfg['python_version'].split(','):
            py_lib_path = os.path.join(lib_dir, 'python%s' % (ver), 'site-packages')
            logging.debug('Installing python module to %s', py_lib_path)
            makedirs(prepend_destdir(py_lib_path))

            py_dir = cfg['python_dir']

            copy_file(os.path.join(py_dir, 'botan3.py'),
                      prepend_destdir(os.path.join(py_lib_path, 'botan3.py')))

    if cfg['with_documentation']:
        target_doc_dir = os.path.join(prefix, cfg['docdir'],
                                      'botan-%d.%d.%d' % (ver_major, ver_minor, ver_patch))

        shutil.rmtree(prepend_destdir(target_doc_dir), True)
        shutil.copytree(cfg['doc_output_dir'], prepend_destdir(target_doc_dir))

        copy_file(os.path.join(cfg['base_dir'], 'license.txt'),
                  prepend_destdir(os.path.join(target_doc_dir, 'license.txt')))
        copy_file(os.path.join(cfg['base_dir'], 'news.rst'),
                  prepend_destdir(os.path.join(target_doc_dir, 'news.txt')))
        for f in [f for f in os.listdir(cfg['doc_dir']) if f.endswith('.txt')]:
            copy_file(os.path.join(cfg['doc_dir'], f), prepend_destdir(os.path.join(target_doc_dir, f)))

        if cfg['with_rst2man']:
            man1_dir = prepend_destdir(os.path.join(prefix, os.path.join(cfg['mandir'], 'man1')))
            makedirs(man1_dir)

            copy_file(os.path.join(cfg['build_dir'], 'botan.1'),
                      os.path.join(man1_dir, 'botan.1'))

    logging.info('Botan %s installation to %s complete', cfg['version'], cfg['prefix'])
    return 0

if __name__ == '__main__':
    try:
        sys.exit(main(sys.argv))
    except Exception as e:
        logging.error('Failure: %s', str(e))
        logging.info(traceback.format_exc())
        sys.exit(1)



#!/usr/bin/env python3

"""
Runs the tests from https://github.com/C2SP/x509-limbo
"""

from botan3 import X509Cert
from dateutil import parser
import json
import datetime
import optparse # pylint: disable=deprecated-module
import re
import subprocess
import sys

ignored_tests = {}

tests_that_succeed_unexpectedly = {
    'rfc5280::aki::critical-aki unexpected': 'Conflates CA and verifier requirements',
    'rfc5280::aki::critical-aki': 'Conflates CA and verifier requirements',
    'rfc5280::aki::intermediate-missing-aki': 'Conflates CA and verifier requirements',
    'rfc5280::aki::leaf-missing-aki': 'Conflates CA and verifier requirements',
    'rfc5280::ee-critical-aia-invalid': 'Conflates CA and verifier requirements',
    'rfc5280::nc::permitted-dns-match-noncritical': 'Conflates CA and verifier requirements',
    'rfc5280::pc::ica-noncritical-pc': 'Conflates CA and verifier requirements',
    'rfc5280::root-non-critical-basic-constraints': 'Conflates CA and verifier requirements',
    'rfc5280::san::noncritical-with-empty-subject': 'Conflates CA and verifier requirements',
    'rfc5280::serial::too-long': 'Conflates CA and verifier requirements',
    'rfc5280::serial::zero': 'Conflates CA and verifier requirements',
    'rfc5280::ski::intermediate-missing-ski': 'Conflates CA and verifier requirements',
    'rfc5280::ski::root-missing-ski': 'Conflates CA and verifier requirements',

    'webpki::aki::root-with-aki-missing-keyidentifier': 'Conflates CA and verifier requirements',
    'webpki::aki::root-with-aki-authoritycertissuer': 'Conflates CA and verifier requirements',
    'webpki::aki::root-with-aki-authoritycertserialnumber': 'Conflates CA and verifier requirements',
    'webpki::aki::root-with-aki-all-fields': 'Conflates CA and verifier requirements',
    'webpki::ee-basicconstraints-ca': 'Conflates CA and verifier requirements',
    'webpki::eku::ee-without-eku': 'Conflates CA and verifier requirements',
    'webpki::san::no-san': 'Conflates CA and verifier requirements',
    'webpki::san::san-critical-with-nonempty-subject': 'Conflates CA and verifier requirements',
    'webpki::v1-cert': 'Conflates CA and verifier requirements',
    'webpki::forbidden-rsa-not-divisable-by-8-in-root': 'Conflates CA and verifier requirements',
    'webpki::forbidden-rsa-key-not-divisable-by-8-in-leaf': 'Conflates CA and verifier requirements',
    'webpki::forbidden-dsa-leaf': 'Conflates CA and verifier requirements',
    'webpki::forbidden-dsa-root': 'Conflates CA and verifier requirements',

    'webpki::forbidden-p192-leaf': 'We do not place restrictions on the leaf key',
    'webpki::forbidden-weak-rsa-in-leaf': 'We do not place restrictions on the leaf key',

    'webpki::san::wildcard-embedded-leftmost-san': 'CABF rule not RFC 5280',
    'webpki::ca-as-leaf': 'Not applicable outside of webpki',

    'webpki::explicit-curve': 'Deprecated but not gone yet',
    'rfc5280::nc::invalid-dnsname-leading-period': 'Common extension',

    'rfc5280::nc::nc-forbids-othername': 'Othername is a NULL which we drop',
    'webpki::san::wildcard-embedded-ulabel-san': 'Needs investigation',
    'webpki::malformed-aia': 'Needs investigation',

    # A number of tests (736, 737, ...) seem to make the implicit assumption
    # that if a name constraint applies to a certificate then we should not
    # ever use the CN as the hostname, even if the ee cert does not have a SAN
    'bettertls::nameconstraints::tc736': 'See comment above',
    'bettertls::nameconstraints::tc737': 'Same as 736',
    'bettertls::nameconstraints::tc738': 'Same as 736',
    'bettertls::nameconstraints::tc742': 'Same as 736',
    'bettertls::nameconstraints::tc743': 'Same as 736',
    'bettertls::nameconstraints::tc744': 'Same as 736',
    'bettertls::nameconstraints::tc745': 'Same as 736',
    'bettertls::nameconstraints::tc746': 'Same as 736',
    'bettertls::nameconstraints::tc747': 'Same as 736',
    'bettertls::nameconstraints::tc751': 'Same as 736',
    'bettertls::nameconstraints::tc752': 'Same as 736',
    'bettertls::nameconstraints::tc753': 'Same as 736',
    'bettertls::nameconstraints::tc754': 'Same as 736',
    'bettertls::nameconstraints::tc755': 'Same as 736',
    'bettertls::nameconstraints::tc756': 'Same as 736',
    'bettertls::nameconstraints::tc760': 'Same as 736',
    'bettertls::nameconstraints::tc761': 'Same as 736',
    'bettertls::nameconstraints::tc762': 'Same as 736',
    'bettertls::nameconstraints::tc763': 'Same as 736',
    'bettertls::nameconstraints::tc764': 'Same as 736',
    'bettertls::nameconstraints::tc765': 'Same as 736',
    'bettertls::nameconstraints::tc769': 'Same as 736',
    'bettertls::nameconstraints::tc770': 'Same as 736',
    'bettertls::nameconstraints::tc771': 'Same as 736',
    'bettertls::nameconstraints::tc772': 'Same as 736',
    'bettertls::nameconstraints::tc773': 'Same as 736',
    'bettertls::nameconstraints::tc774': 'Same as 736',
    'bettertls::nameconstraints::tc778': 'Same as 736',
    'bettertls::nameconstraints::tc779': 'Same as 736',
    'bettertls::nameconstraints::tc780': 'Same as 736',
    'bettertls::nameconstraints::tc781': 'Same as 736',
    'bettertls::nameconstraints::tc782': 'Same as 736',
    'bettertls::nameconstraints::tc783': 'Same as 736',
    'bettertls::nameconstraints::tc787': 'Same as 736',
    'bettertls::nameconstraints::tc788': 'Same as 736',
    'bettertls::nameconstraints::tc789': 'Same as 736',
}

tests_that_fail_unexpectedly = {
    'rfc5280::nc::permitted-ipv6-match': 'IPv6 name constraints not implemented',

    'cve::cve-2024-0567': 'Possible path building bug',
    'rfc5280::root-and-intermediate-swapped': 'Possible path building bug',
    'rfc5280::nc::permitted-self-issued': 'Possible path building bug',
}

def report_success(test_id, modified_result, type):
    if modified_result:
        return "GOOD %s %s (MODIFIED RESULT)" % (test_id, type)
    else:
        return "GOOD %s %s as expected" % (test_id, type)

def report_failure(test_id, modified_result, type):
    if modified_result:
        return "FAIL %s unexpectedly %s (MODIFIED RESULT)" % (test_id, type)
    else:
        return "FAIL %s unexpectedly %s" % (test_id, type)

def dump_x509(who, cert):
    print("%s certificate\n" % (who))

    dump_cmd = ['openssl', 'x509', '-text', '-noout']
    proc = subprocess.run(dump_cmd,
                          input=bytes(cert, 'utf8'),
                          capture_output=True)

    print(proc.stdout.decode('utf8'))

def describe(test):
    print(test['description'])
    dump_x509('Peer', test['peer_certificate'])
    for inter in test['untrusted_intermediates']:
        dump_x509('Intermediate', inter)
    for root in test['trusted_certs']:
        dump_x509('Trusted', root)

def main(args = None):
    if args is None:
        args = sys.argv

    opts = optparse.OptionParser()

    opts.add_option('--run-only', metavar='REGEX', default=None,
                    help='Run only tests matching regex')
    opts.add_option('--stop-on-first-failure', action='store_true', default=False,
                    help='Exit immediately on first failure')

    opts.add_option('--verbose', default=False, action='store_true',
                    help='Print more details')

    (options, args) = opts.parse_args(args)

    if len(args) != 2:
        print("Expected usage: %s <limbo.json>" % (args[0]))
        return 1

    run_only = None
    if options.run_only is not None:
        run_only = re.compile(options.run_only)

    limbo_file = args[1]
    limbo_contents = open(limbo_file).read()
    limbo_json = json.loads(limbo_contents)

    if limbo_json['version'] != 1:
        print("Unexpected version in %s" % (limbo_file))
        return 1

    tests = 0
    success = 0
    ignored = 0
    unexpected_reject = 0
    unexpected_accept = 0
    modified = 0

    for test in limbo_json['testcases']:
        if run_only is not None:
            if run_only.match(test['id']) is None:
                continue

        if test['extended_key_usage'] != [] or test['max_chain_depth'] is not None:
            # we have no way of expressing this here
            ignored += 1
            if options.verbose:
                print("IGNR %s" % (test['id']))
            continue

        if test['id'] in ignored_tests:
            ignored += 1
            continue

        expected_to_pass = test['expected_result'] == 'SUCCESS'
        modified_result = False

        if test['id'] in tests_that_succeed_unexpectedly:
            assert(not expected_to_pass)
            expected_to_pass = True
            modified_result = True
            modified += 1
        elif test['id'] in tests_that_fail_unexpectedly:
            assert(expected_to_pass)
            expected_to_pass = False
            modified_result = True
            modified += 1

        tests += 1

        try:
            trust_roots = [X509Cert(buf=x) for x in test['trusted_certs']]
            intermediates = [X509Cert(buf=x) for x in test['untrusted_intermediates']]
            ee_cert = X509Cert(buf=test['peer_certificate'])
        except Exception as e:
            if not expected_to_pass:
                success += 1
            else:
                print("Test %s unexpected failed at parse time: got %s" % (test['id'], e))
                if options.verbose:
                    print(test)
            continue

        validation_time = 0
        if test['validation_time'] is not None:
            validation_time = int(parser.parse(test['validation_time']).timestamp())

        hostname = None
        if test['expected_peer_name'] != None:
            if test['expected_peer_name']['kind'] in ['DNS', 'IP']:
                hostname = test['expected_peer_name']['value']
            else:
                print("Ignoring peer_name kind %s" % (test['expected_peer_name']['kind']))

        result = ee_cert.verify(intermediates, trust_roots, required_strength=110,
                                hostname=hostname, reference_time=validation_time)

        if result == 0:
            if expected_to_pass:
                success += 1
                if options.verbose:
                    print(report_success(test['id'], modified_result, 'accepted'))
            else:
                unexpected_accept += 1
                print(report_failure(test['id'], modified_result, 'accepted'))
                if options.verbose:
                    describe(test)

                if options.stop_on_first_failure:
                    return 1
        else:
            if not expected_to_pass:
                success += 1
                if options.verbose:
                    print(report_success(test['id'], modified_result, 'rejected'))
            else:
                unexpected_reject += 1
                status_str = X509Cert.validation_status(result)
                print(report_failure(test['id'], modified_result, "rejected with '%s'" % (status_str)))
                if options.verbose:
                    describe(test)
                if options.stop_on_first_failure:
                    return 1

    print("Tests executed: %d" % (tests))
    print("Tests succeeding: %d" % (success))
    print("Tests ignored: %d" % (ignored))
    print("Tests with modified results: %d" % (modified))
    print("Test failures, unexpected accept: %d" % (unexpected_accept))
    print("Test failures, unexpected reject: %d" % (unexpected_reject))

    if unexpected_accept > 0 or unexpected_reject > 0:
        return 1

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

# (C) 2018 Jack Lloyd
# Botan is released under the Simplified BSD License (see license.txt)

# Scans all source and test files and makes sure we are not using a
# BOTAN_HAS_xxx macro which is not actually defined anywhere.

from configure import ModuleInfo, load_info_files
import os
import re

src_dir = 'src'
lib_dir = os.path.join(src_dir, 'lib')

info_modules = load_info_files(lib_dir, 'Modules', "info.txt", ModuleInfo)

all_defines = set()

for module in info_modules.values():
    for define in module._defines:
        all_defines.add(define)

extras = ['MP_DWORD',
          'VALGRIND',
          'ONLINE_REVOCATION_CHECKS',
          'SANITIZER_UNDEFINED',
          'HW_AES_SUPPORT']

for extra in extras:
    all_defines.add(extra)

macro = re.compile('BOTAN_HAS_([A-Za-z0-9_]+)')

for dirname, subdirs, files in os.walk(src_dir):
    for fname in files:
        if fname.endswith('.h') or fname.endswith('.cpp'):
            contents = open(os.path.join(dirname, fname)).read()

            for m in re.finditer(macro, contents):
                if m.group(1) not in all_defines:
                    print('In %s found unknown feature macro %s' % (fname, m.group(1)))




#!/usr/bin/env python3

"""
Implements the "make clean" target

(C) 2017 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import os
import sys
import stat
import re
import optparse  # pylint: disable=deprecated-module
import logging
import json
import shutil
import errno

def remove_dir(d):
    try:
        if os.access(d, os.X_OK):
            logging.debug('Removing directory "%s"', d)
            shutil.rmtree(d)
        else:
            logging.debug('Directory %s was missing', d)
    except Exception as e:
        logging.error('Failed removing directory "%s": %s', d, e)

def remove_file(f):
    try:
        logging.debug('Removing file "%s"', f)
        os.unlink(f)
    except OSError as e:
        if e.errno != errno.ENOENT:
            logging.error('Failed removing file "%s": %s', f, e)

def remove_all_in_dir(d):
    if os.access(d, os.X_OK):
        logging.debug('Removing all files in directory "%s"', d)

        for f in os.listdir(d):
            full_path = os.path.join(d, f)
            mode = os.lstat(full_path).st_mode

            if stat.S_ISDIR(mode):
                remove_dir(full_path)
            else:
                remove_file(full_path)

def parse_options(args):
    parser = optparse.OptionParser()
    parser.add_option('--build-dir', default='build', metavar='DIR',
                      help='specify build dir to clean (default %default)')

    parser.add_option('--distclean', action='store_true', default=False,
                      help='clean everything')
    parser.add_option('--verbose', action='store_true', default=False,
                      help='noisy logging')

    (options, args) = parser.parse_args(args)

    if len(args) > 1:
        raise Exception("Unknown arguments")

    return options

def main(args=None):
    if args is None:
        args = sys.argv

    options = parse_options(args)

    logging.basicConfig(stream=sys.stderr,
                        format='%(levelname) 7s: %(message)s',
                        level=logging.DEBUG if options.verbose else logging.INFO)

    build_dir = options.build_dir

    if not os.access(build_dir, os.X_OK):
        logging.debug('No build directory found')
        # No build dir: clean enough!
        return 0

    build_config_path = os.path.join(build_dir, 'build_config.json')
    build_config_str = None

    try:
        build_config_file = open(build_config_path, encoding='utf8')
        build_config_str = build_config_file.read()
        build_config_file.close()
    except Exception:
        # Ugh have to do generic catch as different exception type thrown in Python2
        logging.error("Unable to access build_config.json in build dir")
        return 1

    build_config = json.loads(build_config_str)

    if options.distclean:
        build_dir = build_config['build_dir']
        remove_file(build_config['makefile_path'])
        remove_dir(build_dir)
    else:
        for dir_type in ['libobj_dir', 'cliobj_dir', 'testobj_dir', 'handbook_output_dir', 'doc_output_dir_doxygen']:
            dir_path = build_config[dir_type]
            if dir_path:
                remove_all_in_dir(dir_path)

        remove_file(build_config['doc_stamp_file'])

    remove_file(build_config['cli_exe'])
    remove_file(build_config['test_exe'])

    lib_basename = build_config['lib_prefix'] + build_config['libname']
    matches_libname = re.compile('^' + lib_basename + '.([a-z]+)((\\.[0-9\\.]+)|$)')

    known_suffix = ['a', 'so', 'dll', 'manifest', 'exp']

    for f in os.listdir(build_config['out_dir']):
        match = matches_libname.match(f)
        if match and match.group(1) in known_suffix:
            remove_file(os.path.join(build_config['out_dir'], f))

    if options.distclean:
        if 'generated_files' in build_config:
            for f in build_config['generated_files'].split(' '):
                remove_file(f)

    return 0

if __name__ == '__main__':
    sys.exit(main())
